from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    BenchmarkProcess,
    BenchmarkProcessApply,
    BenchmarkProcessApplyList,
    BenchmarkProcessList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class BenchmarkProcessProductionPlanTimeSeriesAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "BenchmarkProcess.productionPlanTimeSeries"},
        )
        if isinstance(external_id, str):
            is_benchmark_proces = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_benchmark_proces)
            )

        else:
            is_benchmark_process = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_benchmark_process)
            )

    def list(self, benchmark_proces_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "BenchmarkProcess.productionPlanTimeSeries"},
        )
        filters.append(is_edge_type)
        if benchmark_proces_id:
            benchmark_proces_ids = (
                [benchmark_proces_id] if isinstance(benchmark_proces_id, str) else benchmark_proces_id
            )
            is_benchmark_process = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in benchmark_proces_ids],
            )
            filters.append(is_benchmark_process)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class BenchmarkProcessAPI(TypeAPI[BenchmarkProcess, BenchmarkProcessApply, BenchmarkProcessList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BenchmarkProcess,
            class_apply_type=BenchmarkProcessApply,
            class_list=BenchmarkProcessList,
        )
        self.view_id = view_id
        self.production_plan_time_series = BenchmarkProcessProductionPlanTimeSeriesAPI(client)

    def apply(
        self, benchmark_proces: BenchmarkProcessApply | Sequence[BenchmarkProcessApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(benchmark_proces, BenchmarkProcessApply):
            instances = benchmark_proces.to_instances_apply()
        else:
            instances = BenchmarkProcessApplyList(benchmark_proces).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(BenchmarkProcessApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(BenchmarkProcessApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> BenchmarkProcess:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> BenchmarkProcessList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> BenchmarkProcess | BenchmarkProcessList:
        if isinstance(external_id, str):
            benchmark_proces = self._retrieve((self.sources.space, external_id))

            production_plan_time_series_edges = self.production_plan_time_series.retrieve(external_id)
            benchmark_proces.production_plan_time_series = [
                edge.end_node.external_id for edge in production_plan_time_series_edges
            ]

            return benchmark_proces
        else:
            benchmark_process = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            production_plan_time_series_edges = self.production_plan_time_series.retrieve(external_id)
            self._set_production_plan_time_series(benchmark_process, production_plan_time_series_edges)

            return benchmark_process

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> BenchmarkProcessList:
        filter_ = _create_filter(
            self.view_id,
            name,
            name_prefix,
            external_id_prefix,
            filter,
        )

        benchmark_process = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            production_plan_time_series_edges = self.production_plan_time_series.list(
                benchmark_process.as_external_ids(), limit=-1
            )
            self._set_production_plan_time_series(benchmark_process, production_plan_time_series_edges)

        return benchmark_process

    @staticmethod
    def _set_production_plan_time_series(
        benchmark_process: Sequence[BenchmarkProcess], production_plan_time_series_edges: Sequence[dm.Edge]
    ):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in production_plan_time_series_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for benchmark_proces in benchmark_process:
            node_id = benchmark_proces.id_tuple()
            if node_id in edges_by_start_node:
                benchmark_proces.production_plan_time_series = [
                    edge.end_node.external_id for edge in edges_by_start_node[node_id]
                ]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    Watercourse,
    WatercourseApply,
    WatercourseApplyList,
    WatercourseList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class WatercoursePlantsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Watercourse.plants"},
        )
        if isinstance(external_id, str):
            is_watercourse = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_watercourse)
            )

        else:
            is_watercourses = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_watercourses)
            )

    def list(self, watercourse_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Watercourse.plants"},
        )
        filters.append(is_edge_type)
        if watercourse_id:
            watercourse_ids = [watercourse_id] if isinstance(watercourse_id, str) else watercourse_id
            is_watercourses = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in watercourse_ids],
            )
            filters.append(is_watercourses)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class WatercourseAPI(TypeAPI[Watercourse, WatercourseApply, WatercourseList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Watercourse,
            class_apply_type=WatercourseApply,
            class_list=WatercourseList,
        )
        self.view_id = view_id
        self.plants = WatercoursePlantsAPI(client)

    def apply(
        self, watercourse: WatercourseApply | Sequence[WatercourseApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(watercourse, WatercourseApply):
            instances = watercourse.to_instances_apply()
        else:
            instances = WatercourseApplyList(watercourse).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(WatercourseApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(WatercourseApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Watercourse:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> WatercourseList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Watercourse | WatercourseList:
        if isinstance(external_id, str):
            watercourse = self._retrieve((self.sources.space, external_id))

            plant_edges = self.plants.retrieve(external_id)
            watercourse.plants = [edge.end_node.external_id for edge in plant_edges]

            return watercourse
        else:
            watercourses = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            plant_edges = self.plants.retrieve(external_id)
            self._set_plants(watercourses, plant_edges)

            return watercourses

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> WatercourseList:
        filter_ = _create_filter(
            self.view_id,
            name,
            name_prefix,
            external_id_prefix,
            filter,
        )

        watercourses = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            plant_edges = self.plants.list(watercourses.as_external_ids(), limit=-1)
            self._set_plants(watercourses, plant_edges)

        return watercourses

    @staticmethod
    def _set_plants(watercourses: Sequence[Watercourse], plant_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in plant_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for watercourse in watercourses:
            node_id = watercourse.id_tuple()
            if node_id in edges_by_start_node:
                watercourse.plants = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    OutputMapping,
    OutputMappingApply,
    OutputMappingApplyList,
    OutputMappingList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class OutputMappingAPI(TypeAPI[OutputMapping, OutputMappingApply, OutputMappingList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=OutputMapping,
            class_apply_type=OutputMappingApply,
            class_list=OutputMappingList,
        )
        self.view_id = view_id

    def apply(
        self, output_mapping: OutputMappingApply | Sequence[OutputMappingApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(output_mapping, OutputMappingApply):
            instances = output_mapping.to_instances_apply()
        else:
            instances = OutputMappingApplyList(output_mapping).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(OutputMappingApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(OutputMappingApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> OutputMapping:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> OutputMappingList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> OutputMapping | OutputMappingList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        shop_object_type: str | list[str] | None = None,
        shop_object_type_prefix: str | None = None,
        shop_attribute_name: str | list[str] | None = None,
        shop_attribute_name_prefix: str | None = None,
        cdf_attribute_name: str | list[str] | None = None,
        cdf_attribute_name_prefix: str | None = None,
        unit: str | list[str] | None = None,
        unit_prefix: str | None = None,
        is_step: bool | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> OutputMappingList:
        filter_ = _create_filter(
            self.view_id,
            shop_object_type,
            shop_object_type_prefix,
            shop_attribute_name,
            shop_attribute_name_prefix,
            cdf_attribute_name,
            cdf_attribute_name_prefix,
            unit,
            unit_prefix,
            is_step,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    shop_object_type: str | list[str] | None = None,
    shop_object_type_prefix: str | None = None,
    shop_attribute_name: str | list[str] | None = None,
    shop_attribute_name_prefix: str | None = None,
    cdf_attribute_name: str | list[str] | None = None,
    cdf_attribute_name_prefix: str | None = None,
    unit: str | list[str] | None = None,
    unit_prefix: str | None = None,
    is_step: bool | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if shop_object_type and isinstance(shop_object_type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopObjectType"), value=shop_object_type))
    if shop_object_type and isinstance(shop_object_type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopObjectType"), values=shop_object_type))
    if shop_object_type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopObjectType"), value=shop_object_type_prefix))
    if shop_attribute_name and isinstance(shop_attribute_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopAttributeName"), value=shop_attribute_name))
    if shop_attribute_name and isinstance(shop_attribute_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopAttributeName"), values=shop_attribute_name))
    if shop_attribute_name_prefix:
        filters.append(
            dm.filters.Prefix(view_id.as_property_ref("shopAttributeName"), value=shop_attribute_name_prefix)
        )
    if cdf_attribute_name and isinstance(cdf_attribute_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("cdfAttributeName"), value=cdf_attribute_name))
    if cdf_attribute_name and isinstance(cdf_attribute_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("cdfAttributeName"), values=cdf_attribute_name))
    if cdf_attribute_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("cdfAttributeName"), value=cdf_attribute_name_prefix))
    if unit and isinstance(unit, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("unit"), value=unit))
    if unit and isinstance(unit, list):
        filters.append(dm.filters.In(view_id.as_property_ref("unit"), values=unit))
    if unit_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("unit"), value=unit_prefix))
    if is_step and isinstance(is_step, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("isStep"), value=is_step))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    RKOMCombinationBid,
    RKOMCombinationBidApply,
    RKOMCombinationBidApplyList,
    RKOMCombinationBidList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class RKOMCombinationBidAPI(TypeAPI[RKOMCombinationBid, RKOMCombinationBidApply, RKOMCombinationBidList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=RKOMCombinationBid,
            class_apply_type=RKOMCombinationBidApply,
            class_list=RKOMCombinationBidList,
        )
        self.view_id = view_id

    def apply(
        self, rkom_combination_bid: RKOMCombinationBidApply | Sequence[RKOMCombinationBidApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(rkom_combination_bid, RKOMCombinationBidApply):
            instances = rkom_combination_bid.to_instances_apply()
        else:
            instances = RKOMCombinationBidApplyList(rkom_combination_bid).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(RKOMCombinationBidApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(RKOMCombinationBidApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> RKOMCombinationBid:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> RKOMCombinationBidList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> RKOMCombinationBid | RKOMCombinationBidList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        auction: str | list[str] | None = None,
        auction_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> RKOMCombinationBidList:
        filter_ = _create_filter(
            self.view_id,
            name,
            name_prefix,
            auction,
            auction_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    auction: str | list[str] | None = None,
    auction_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if auction and isinstance(auction, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("auction"), value=auction))
    if auction and isinstance(auction, list):
        filters.append(dm.filters.In(view_id.as_property_ref("auction"), values=auction))
    if auction_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("auction"), value=auction_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    MarketAgreement,
    MarketAgreementApply,
    MarketAgreementApplyList,
    MarketAgreementList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class MarketAgreementAPI(TypeAPI[MarketAgreement, MarketAgreementApply, MarketAgreementList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=MarketAgreement,
            class_apply_type=MarketAgreementApply,
            class_list=MarketAgreementList,
        )
        self.view_id = view_id

    def apply(
        self, market_agreement: MarketAgreementApply | Sequence[MarketAgreementApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(market_agreement, MarketAgreementApply):
            instances = market_agreement.to_instances_apply()
        else:
            instances = MarketAgreementApplyList(market_agreement).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(MarketAgreementApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(MarketAgreementApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> MarketAgreement:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> MarketAgreementList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> MarketAgreement | MarketAgreementList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        m_rid: str | list[str] | None = None,
        m_rid_prefix: str | None = None,
        type: str | list[str] | None = None,
        type_prefix: str | None = None,
        min_created_timestamp: datetime.datetime | None = None,
        max_created_timestamp: datetime.datetime | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> MarketAgreementList:
        filter_ = _create_filter(
            self.view_id,
            m_rid,
            m_rid_prefix,
            type,
            type_prefix,
            min_created_timestamp,
            max_created_timestamp,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    m_rid: str | list[str] | None = None,
    m_rid_prefix: str | None = None,
    type: str | list[str] | None = None,
    type_prefix: str | None = None,
    min_created_timestamp: datetime.datetime | None = None,
    max_created_timestamp: datetime.datetime | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if m_rid and isinstance(m_rid, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("mRID"), value=m_rid))
    if m_rid and isinstance(m_rid, list):
        filters.append(dm.filters.In(view_id.as_property_ref("mRID"), values=m_rid))
    if m_rid_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("mRID"), value=m_rid_prefix))
    if type and isinstance(type, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("type"), value=type))
    if type and isinstance(type, list):
        filters.append(dm.filters.In(view_id.as_property_ref("type"), values=type))
    if type_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("type"), value=type_prefix))
    if min_created_timestamp or max_created_timestamp:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("createdTimestamp"),
                gte=min_created_timestamp.isoformat() if min_created_timestamp else None,
                lte=max_created_timestamp.isoformat() if max_created_timestamp else None,
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, TypeApplyList, TypeList

if TYPE_CHECKING:
    from ._value_transformation import ValueTransformationApply

__all__ = [
    "InputTimeSeriesMapping",
    "InputTimeSeriesMappingApply",
    "InputTimeSeriesMappingList",
    "InputTimeSeriesMappingApplyList",
]


class InputTimeSeriesMapping(DomainModel):
    space: ClassVar[str] = "power-ops"
    shop_object_type: Optional[str] = Field(None, alias="shopObjectType")
    shop_object_name: Optional[str] = Field(None, alias="shopObjectName")
    shop_attribute_name: Optional[str] = Field(None, alias="shopAttributeName")
    cdf_time_series: Optional[str] = Field(None, alias="cdfTimeSeries")
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None
    transformations: Optional[list[str]] = None

    def as_apply(self) -> InputTimeSeriesMappingApply:
        return InputTimeSeriesMappingApply(
            external_id=self.external_id,
            shop_object_type=self.shop_object_type,
            shop_object_name=self.shop_object_name,
            shop_attribute_name=self.shop_attribute_name,
            cdf_time_series=self.cdf_time_series,
            retrieve=self.retrieve,
            aggregation=self.aggregation,
            transformations=self.transformations,
        )


class InputTimeSeriesMappingApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    shop_object_type: Optional[str] = None
    shop_object_name: Optional[str] = None
    shop_attribute_name: Optional[str] = None
    cdf_time_series: Optional[str] = None
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None
    transformations: Union[list[ValueTransformationApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> dm.InstancesApply:
        if self.external_id in cache:
            return dm.InstancesApply(dm.NodeApplyList([]), dm.EdgeApplyList([]))

        sources = []
        properties = {}
        if self.shop_object_type is not None:
            properties["shopObjectType"] = self.shop_object_type
        if self.shop_object_name is not None:
            properties["shopObjectName"] = self.shop_object_name
        if self.shop_attribute_name is not None:
            properties["shopAttributeName"] = self.shop_attribute_name
        if self.cdf_time_series is not None:
            properties["cdfTimeSeries"] = self.cdf_time_series
        if self.retrieve is not None:
            properties["retrieve"] = self.retrieve
        if self.aggregation is not None:
            properties["aggregation"] = self.aggregation
        if properties:
            source = dm.NodeOrEdgeData(
                source=dm.ContainerId("power-ops", "InputTimeSeriesMapping"),
                properties=properties,
            )
            sources.append(source)
        if sources:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                sources=sources,
            )
            nodes = [this_node]
        else:
            nodes = []

        edges = []
        cache.add(self.external_id)

        for transformation in self.transformations or []:
            edge = self._create_transformation_edge(transformation)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(transformation, DomainModelApply):
                instances = transformation._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return dm.InstancesApply(dm.NodeApplyList(nodes), dm.EdgeApplyList(edges))

    def _create_transformation_edge(self, transformation: Union[str, ValueTransformationApply]) -> dm.EdgeApply:
        if isinstance(transformation, str):
            end_node_ext_id = transformation
        elif isinstance(transformation, DomainModelApply):
            end_node_ext_id = transformation.external_id
        else:
            raise TypeError(f"Expected str or ValueTransformationApply, got {type(transformation)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "InputTimeSeriesMapping.transformations"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class InputTimeSeriesMappingList(TypeList[InputTimeSeriesMapping]):
    _NODE = InputTimeSeriesMapping

    def as_apply(self) -> InputTimeSeriesMappingApplyList:
        return InputTimeSeriesMappingApplyList([node.as_apply() for node in self.data])


class InputTimeSeriesMappingApplyList(TypeApplyList[InputTimeSeriesMappingApply]):
    _NODE = InputTimeSeriesMappingApply



from __future__ import annotations

from collections import defaultdict
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import (
    DayAheadBid,
    DayAheadBidApply,
    DayAheadBidApplyList,
    DayAheadBidList,
)

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class DayAheadBidDateAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadBid.date"},
        )
        if isinstance(external_id, str):
            is_day_ahead_bid = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_bid)
            )

        else:
            is_day_ahead_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_bids)
            )

    def list(self, day_ahead_bid_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadBid.date"},
        )
        filters.append(is_edge_type)
        if day_ahead_bid_id:
            day_ahead_bid_ids = [day_ahead_bid_id] if isinstance(day_ahead_bid_id, str) else day_ahead_bid_id
            is_day_ahead_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in day_ahead_bid_ids],
            )
            filters.append(is_day_ahead_bids)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class DayAheadBidPriceScenariosAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadBid.priceScenarios"},
        )
        if isinstance(external_id, str):
            is_day_ahead_bid = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_bid)
            )

        else:
            is_day_ahead_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_day_ahead_bids)
            )

    def list(self, day_ahead_bid_id: str | list[str] | None = None, limit=DEFAULT_LIMIT_READ) -> dm.EdgeList:
        f = dm.filters
        filters = []
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "DayAheadBid.priceScenarios"},
        )
        filters.append(is_edge_type)
        if day_ahead_bid_id:
            day_ahead_bid_ids = [day_ahead_bid_id] if isinstance(day_ahead_bid_id, str) else day_ahead_bid_id
            is_day_ahead_bids = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in day_ahead_bid_ids],
            )
            filters.append(is_day_ahead_bids)

        return self._client.data_modeling.instances.list("edge", limit=limit, filter=f.And(*filters))


class DayAheadBidAPI(TypeAPI[DayAheadBid, DayAheadBidApply, DayAheadBidList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=DayAheadBid,
            class_apply_type=DayAheadBidApply,
            class_list=DayAheadBidList,
        )
        self.view_id = view_id
        self.date = DayAheadBidDateAPI(client)
        self.price_scenarios = DayAheadBidPriceScenariosAPI(client)

    def apply(
        self, day_ahead_bid: DayAheadBidApply | Sequence[DayAheadBidApply], replace: bool = False
    ) -> dm.InstancesApplyResult:
        if isinstance(day_ahead_bid, DayAheadBidApply):
            instances = day_ahead_bid.to_instances_apply()
        else:
            instances = DayAheadBidApplyList(day_ahead_bid).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(DayAheadBidApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(DayAheadBidApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> DayAheadBid:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> DayAheadBidList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> DayAheadBid | DayAheadBidList:
        if isinstance(external_id, str):
            day_ahead_bid = self._retrieve((self.sources.space, external_id))

            date_edges = self.date.retrieve(external_id)
            day_ahead_bid.date = [edge.end_node.external_id for edge in date_edges]
            price_scenario_edges = self.price_scenarios.retrieve(external_id)
            day_ahead_bid.price_scenarios = [edge.end_node.external_id for edge in price_scenario_edges]

            return day_ahead_bid
        else:
            day_ahead_bids = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            date_edges = self.date.retrieve(external_id)
            self._set_date(day_ahead_bids, date_edges)
            price_scenario_edges = self.price_scenarios.retrieve(external_id)
            self._set_price_scenarios(day_ahead_bids, price_scenario_edges)

            return day_ahead_bids

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        is_default_config_for_price_area: bool | None = None,
        main_scenario: str | list[str] | None = None,
        main_scenario_prefix: str | None = None,
        price_area: str | list[str] | None = None,
        price_area_prefix: str | None = None,
        watercourse: str | list[str] | None = None,
        watercourse_prefix: str | None = None,
        no_shop: bool | None = None,
        bid_process_configuration_name: str | list[str] | None = None,
        bid_process_configuration_name_prefix: str | None = None,
        bid_matrix_generator_config_external_id: str | list[str] | None = None,
        bid_matrix_generator_config_external_id_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> DayAheadBidList:
        filter_ = _create_filter(
            self.view_id,
            name,
            name_prefix,
            is_default_config_for_price_area,
            main_scenario,
            main_scenario_prefix,
            price_area,
            price_area_prefix,
            watercourse,
            watercourse_prefix,
            no_shop,
            bid_process_configuration_name,
            bid_process_configuration_name_prefix,
            bid_matrix_generator_config_external_id,
            bid_matrix_generator_config_external_id_prefix,
            external_id_prefix,
            filter,
        )

        day_ahead_bids = self._list(limit=limit, filter=filter_)

        if retrieve_edges:
            date_edges = self.date.list(day_ahead_bids.as_external_ids(), limit=-1)
            self._set_date(day_ahead_bids, date_edges)
            price_scenario_edges = self.price_scenarios.list(day_ahead_bids.as_external_ids(), limit=-1)
            self._set_price_scenarios(day_ahead_bids, price_scenario_edges)

        return day_ahead_bids

    @staticmethod
    def _set_date(day_ahead_bids: Sequence[DayAheadBid], date_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in date_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for day_ahead_bid in day_ahead_bids:
            node_id = day_ahead_bid.id_tuple()
            if node_id in edges_by_start_node:
                day_ahead_bid.date = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_price_scenarios(day_ahead_bids: Sequence[DayAheadBid], price_scenario_edges: Sequence[dm.Edge]):
        edges_by_start_node: dict[tuple, list] = defaultdict(list)
        for edge in price_scenario_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for day_ahead_bid in day_ahead_bids:
            node_id = day_ahead_bid.id_tuple()
            if node_id in edges_by_start_node:
                day_ahead_bid.price_scenarios = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]


def _create_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    is_default_config_for_price_area: bool | None = None,
    main_scenario: str | list[str] | None = None,
    main_scenario_prefix: str | None = None,
    price_area: str | list[str] | None = None,
    price_area_prefix: str | None = None,
    watercourse: str | list[str] | None = None,
    watercourse_prefix: str | None = None,
    no_shop: bool | None = None,
    bid_process_configuration_name: str | list[str] | None = None,
    bid_process_configuration_name_prefix: str | None = None,
    bid_matrix_generator_config_external_id: str | list[str] | None = None,
    bid_matrix_generator_config_external_id_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if is_default_config_for_price_area and isinstance(is_default_config_for_price_area, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("isDefaultConfigForPriceArea"), value=is_default_config_for_price_area
            )
        )
    if main_scenario and isinstance(main_scenario, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("mainScenario"), value=main_scenario))
    if main_scenario and isinstance(main_scenario, list):
        filters.append(dm.filters.In(view_id.as_property_ref("mainScenario"), values=main_scenario))
    if main_scenario_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("mainScenario"), value=main_scenario_prefix))
    if price_area and isinstance(price_area, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("priceArea"), value=price_area))
    if price_area and isinstance(price_area, list):
        filters.append(dm.filters.In(view_id.as_property_ref("priceArea"), values=price_area))
    if price_area_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("priceArea"), value=price_area_prefix))
    if watercourse and isinstance(watercourse, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("watercourse"), value=watercourse))
    if watercourse and isinstance(watercourse, list):
        filters.append(dm.filters.In(view_id.as_property_ref("watercourse"), values=watercourse))
    if watercourse_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("watercourse"), value=watercourse_prefix))
    if no_shop and isinstance(no_shop, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("noShop"), value=no_shop))
    if bid_process_configuration_name and isinstance(bid_process_configuration_name, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidProcessConfigurationName"), value=bid_process_configuration_name
            )
        )
    if bid_process_configuration_name and isinstance(bid_process_configuration_name, list):
        filters.append(
            dm.filters.In(view_id.as_property_ref("bidProcessConfigurationName"), values=bid_process_configuration_name)
        )
    if bid_process_configuration_name_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("bidProcessConfigurationName"), value=bid_process_configuration_name_prefix
            )
        )
    if bid_matrix_generator_config_external_id and isinstance(bid_matrix_generator_config_external_id, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidMatrixGeneratorConfigExternalId"),
                value=bid_matrix_generator_config_external_id,
            )
        )
    if bid_matrix_generator_config_external_id and isinstance(bid_matrix_generator_config_external_id, list):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidMatrixGeneratorConfigExternalId"),
                values=bid_matrix_generator_config_external_id,
            )
        )
    if bid_matrix_generator_config_external_id_prefix:
        filters.append(
            dm.filters.Prefix(
                view_id.as_property_ref("bidMatrixGeneratorConfigExternalId"),
                value=bid_matrix_generator_config_external_id_prefix,
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.powerops.client._generated.data_classes import Reason, ReasonApply, ReasonApplyList, ReasonList

from ._core import DEFAULT_LIMIT_READ, TypeAPI


class ReasonAPI(TypeAPI[Reason, ReasonApply, ReasonList]):
    def __init__(self, client: CogniteClient, view_id: dm.ViewId):
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Reason,
            class_apply_type=ReasonApply,
            class_list=ReasonList,
        )
        self.view_id = view_id

    def apply(self, reason: ReasonApply | Sequence[ReasonApply], replace: bool = False) -> dm.InstancesApplyResult:
        if isinstance(reason, ReasonApply):
            instances = reason.to_instances_apply()
        else:
            instances = ReasonApplyList(reason).to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(ReasonApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(ReasonApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Reason:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> ReasonList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Reason | ReasonList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(
        self,
        code: str | list[str] | None = None,
        code_prefix: str | None = None,
        text: str | list[str] | None = None,
        text_prefix: str | None = None,
        external_id_prefix: str | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> ReasonList:
        filter_ = _create_filter(
            self.view_id,
            code,
            code_prefix,
            text,
            text_prefix,
            external_id_prefix,
            filter,
        )

        return self._list(limit=limit, filter=filter_)


def _create_filter(
    view_id: dm.ViewId,
    code: str | list[str] | None = None,
    code_prefix: str | None = None,
    text: str | list[str] | None = None,
    text_prefix: str | None = None,
    external_id_prefix: str | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if code and isinstance(code, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("code"), value=code))
    if code and isinstance(code, list):
        filters.append(dm.filters.In(view_id.as_property_ref("code"), values=code))
    if code_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("code"), value=code_prefix))
    if text and isinstance(text, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("text"), value=text))
    if text and isinstance(text, list):
        filters.append(dm.filters.In(view_id.as_property_ref("text"), values=text))
    if text_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("text"), value=text_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None
